@@ -1,17 +1,21 @@
 ## integration tests for Python client
 ## to run tests: nosetests -vs
-## to run single test: nosetests -vs synapse/integration_test.py:TestClient.test_downloadFile
+## to run single test: nosetests -vs synapseclient/integration_test.py:TestClient.test_downloadFile
 
 import client, utils
+from client import Activity
 from version_check import version_check
 import ConfigParser
 from nose.tools import *
 import tempfile
 import os
+import ConfigParser
+from datetime import datetime
 
 
 PROJECT_JSON={ u'entityType': u'org.sagebionetworks.repo.model.Project', u'name': ''}
 DATA_JSON={ u'entityType': u'org.sagebionetworks.repo.model.Data', u'parentId': ''}
+CODE_JSON={ u'entityType': u'org.sagebionetworks.repo.model.Code', u'parentId': ''}
 
 
 class TestClient:
@@ -46,18 +50,6 @@ class TestClient:
         self.toRemove.append(entity)
         return entity
 
-    def test__connect(self):
-        import httplib
-        #Test https protocol
-        self.syn.repoEndpoint['protocol']='https'
-        conn = self.syn._connect(self.syn.repoEndpoint)
-        assert isinstance(conn, httplib.HTTPSConnection)
-
-        #Test http protocol
-        self.syn.repoEndpoint['protocol']='http'
-        conn = self.syn._connect(self.syn.repoEndpoint)
-        assert isinstance(conn, httplib.HTTPConnection)
-
 
     def test_printEntity(self):
         self.syn.printEntity({'hello':'world', 'alist':[1,2,3,4]}) 
@@ -72,7 +64,6 @@ class TestClient:
         self.syn.login()
 
         #Test that it works with username and password
-        import ConfigParser
         config = ConfigParser.ConfigParser()
         config.read(client.CONFIG_FILE)
         self.syn.login(config.get('authentication', 'username'), config.get('authentication', 'password'))
@@ -86,7 +77,6 @@ class TestClient:
         os.remove(os.path.join(client.CACHE_DIR, '.session'))
         assert_raises(Exception, self.syn.login)
 
-        
 
     def test_getEntity(self):
         #Create a new project
@@ -99,9 +89,6 @@ class TestClient:
         returnEntity = self.syn.getEntity(entity['id'])
         assert entity == returnEntity
 
-        #Check that I can get annotations
-        print self.syn.getEntity(returnEntity['annotations'])
-
 
     def test_loadEntity(self):
         #loadEntity does the same thing as downloadEntity so nothing new to test
@@ -123,6 +110,10 @@ class TestClient:
         returnEntity = self.syn.getEntity(entity['id'])
         assert entity == returnEntity
 
+        #Create entity with provenance record
+        self.syn.deleteEntity(returnEntity['id'])
+        entity = self.syn.createEntity(DATA_JSON, used='syn123')
+
 
     def test_updateEntity(self):
         entity = self.createProject()
@@ -174,12 +165,15 @@ class TestClient:
         entity = self.syn.createEntity(DATA_JSON)       
 
         #create a temporary file
-        (fd, fname) = tempfile.mkstemp()
+        (fp, fname) = tempfile.mkstemp()
+        with open(fname, 'w') as f:
+            f.write('foo\n')
         self.syn.uploadFile(entity, fname)
 
         #Download and verify that it is the same filename
         entity = self.syn.downloadEntity(entity)
         assert entity['files'][0]==os.path.split(fname)[-1]
+        os.remove(fname)
 
 
     def test_downloadFile(self):
@@ -219,6 +213,112 @@ class TestClient:
         assert not version_check(current_version="999.999.999", version_url="http://dev-versions.synapse.sagebase.org/bad_filename_doesnt_exist")
 
 
+    def test_provenance(self):
+        """
+        test provenance features
+        """
+
+        ## create a new project
+        project = self.createProject()
+
+        # make some bogus data
+        import random
+        random.seed(12345)
+        data = [random.gauss(mu=0.0, sigma=1.0) for i in range(100)]
+
+        ## create a data entity
+        try:
+            try:
+                f = tempfile.NamedTemporaryFile(suffix=".txt", delete=False)
+                f.write(", ".join((str(n) for n in data)))
+                f.write("\n")
+            finally:
+                f.close()
+            DATA_JSON['parentId']= project['id']
+            data_entity = self.syn.createEntity(DATA_JSON)
+            data_entity = self.syn.uploadFile(data_entity, f.name)
+        finally:
+            os.remove(f.name)
+
+        ## create a code entity, source of the data above
+        code = """
+        ## Chris's fabulous random data generator
+        ############################################################
+        import random
+        random.seed(12345)
+        data = [random.gauss(mu=0.0, sigma=1.0) for i in range(100)]
+        """
+        try:
+            try:
+                f = tempfile.NamedTemporaryFile(suffix=".py", delete=False)
+                f.write(code)
+                f.write("\n")
+            finally:
+                f.close()
+            CODE_JSON['parentId']= project['id']
+            code_entity = self.syn.createEntity(CODE_JSON)
+            code_entity = self.syn.uploadFile(code_entity, f.name)
+        finally:
+            os.remove(f.name)
+
+        ## create a new activity asserting that the code entity was used to
+        ## make the data entity
+        activity = Activity(name='random.gauss', description='Generate some random numbers')
+        activity.used(code_entity['id'], wasExecuted=True)
+
+        activity = self.syn.setProvenance(data_entity, activity)
+
+        ## retrieve the saved provenance record
+        retrieved_activity = self.syn.getProvenance(data_entity['id'])
+
+        ## does it match what we expect?
+        assert retrieved_activity == activity
+
+        ## test update
+        new_description = 'Generate random numbers like a gangsta'
+        retrieved_activity['description'] = new_description
+        updated_activity = self.syn.updateActivity(retrieved_activity)
+
+        ## did the update stick?
+        assert updated_activity['name'] == retrieved_activity['name']
+        assert updated_activity['description'] == new_description
+
+        ## test delete
+        self.syn.deleteProvenance(data_entity)
+
+        ## should be gone now
+        deleted_activity = self.syn.getProvenance(data_entity['id'])
+        assert deleted_activity is None
+
+
+    def test_annotations(self):
+        ## create a new project
+        project = self.createProject()
+
+        DATA_JSON['parentId']= project['id']
+        entity = self.syn.createEntity(DATA_JSON)
+
+        a = self.syn._getAnnotations(entity)
+        assert 'etag' in a
+
+        a['bogosity'] = 'total'
+        self.syn._setAnnotations(entity, a)
+
+        a2 = self.syn._getAnnotations(entity)
+        assert a2['bogosity'] == ['total']
+
+        a2['primes'] = [2,3,5,7,11,13,17,19,23,29]
+        a2['phat_numbers'] = [1234.5678, 8888.3333, 1212.3434, 6677.8899]
+        a2['goobers'] = ['chris', 'jen', 'jane']
+        a2['present_time'] = datetime.now()
+
+        self.syn._setAnnotations(entity, a2)
+        a3 = self.syn._getAnnotations(entity)
+        assert a3['primes'] == [2,3,5,7,11,13,17,19,23,29]
+        assert a3['phat_numbers'] == [1234.5678, 8888.3333, 1212.3434, 6677.8899]
+        assert a3['goobers'] == ['chris', 'jen', 'jane']
+        ## only accurate to within a second 'cause synapse strips off the fractional part
+        assert a3['present_time'][0].strftime('%Y-%m-%d %H:%M:%S') == a2['present_time'].strftime('%Y-%m-%d %H:%M:%S')
 
 
 
@@ -233,3 +333,5 @@ if __name__ == '__main__':
     test.test_query()
     test.test_uploadFile()
     test.test_version_check()
+    test.test_provenance()
+
